/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000 NETSYS GmbH. All Rights Reserved.
//
// MODULE: 2os2_lnx.c 
//
// DESCRIPTION:
//              Linux kernel driver for VM execution
//
// REMARK:
//              Driver is the main instrument to transmit control and 
//              informational messages between Application and Monitor. 
//              Driver implements functions for preparation of Monitor
//              execution environment. 
//
/////////////////////////////////////////////////////////////////////////////

#include <linux/autoconf.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0) 
 #error "You need at least 2.4.0 kernel to compile that module"
#endif

#if CONFIG_X86_PAE
 #error "You have PAE enabled in your kernel. twoOStwo do not support this configuration"
#endif

#define EXPORT_SYMTAB
#ifdef CONFIG_MODVERSIONS
 #define MODVERSIONS
        #include <linux/modversions.h>
#endif
#ifndef __KERNEL__
        #define __KERNEL__
#endif

#ifndef VERSION_CODE
       #define VERSION_CODE(vers, rel, seq) (((vers) << 16) | ((rel) << 8) | (seq))
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/slab.h>

#include <asm/uaccess.h>
#ifndef __PAGE_OFFSET
     #define __PAGE_OFFSET (0xc0000000)
#endif
#define KERNEL_OFFSET __PAGE_OFFSET

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif

#include <asm/unistd.h>
#include <asm/io.h>

#include "vmTest.h"
#include "page_man.h"

/////////////////////////////////////////////////////////////////////////////
//
// Constants which required for Linux Driver execution
// (from kernel version 2.1.18)  
//
/////////////////////////////////////////////////////////////////////////////
#if !defined(THIS_MODULE)
       #define THIS_MODULE (&__this_module)
#endif

// Module dynamically allocates its major number
// vmTest_major stores driver major number
#define VMTEST_MAJOR 0
int vmTest_major = VMTEST_MAJOR;

MODULE_PARM(vmTest_major, "i");
MODULE_PARM_DESC(vmTest_major, "major number (default " __MODULE_STRING(VMTEST_MAJOR) ")");


// function of the interaction with Application (allows to exchange requests)  
static int vmTest_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

// processing of open file event  
static int vmTest_open(struct inode *, struct file *);

// mapping Application memory
static int vmTest_mmap(struct file * file, struct vm_area_struct * vma);

// processing of close file event  
static int vmTest_close(struct inode *, struct file *);

void di_printk(char *fmt, ...);

/////////////////////////////////////////////////////////////////////////////
//
// STRUCTURE: vmTest_fops
//
// DESCRIPTION:
//              Structure defines functions that can be used to access
//              Driver (entry points of Driver)
//
/////////////////////////////////////////////////////////////////////////////
static struct file_operations vmTest_fops = 
{
        owner:    THIS_MODULE,
        ioctl:    vmTest_ioctl,
        open:     vmTest_open,
        release:  vmTest_close,
        mmap:     vmTest_mmap,
};


/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: init_module
//
// DESCRIPTION:
//              Driver setting and initialization
//
// PARAMETERS:
//
// RETURN VALUE:
//              If the function succeeds, the return value is 0 
//
/////////////////////////////////////////////////////////////////////////////
int init_module(void)
{
        int iRetVal;

        di_printk("vmTest: init module...");

        // Registrate Driver as character device
        iRetVal = register_chrdev(vmTest_major, "vmTest", &vmTest_fops);
        if (iRetVal < 0)
        {
                di_printk("vmTest: register_chrdev() error %d", iRetVal);
                return iRetVal;
        }

        // Store driver major number
        if(!vmTest_major)
                vmTest_major = iRetVal;

        return 0;
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: cleanup_module
//
// DESCRIPTION:
//              Driver deinitialization and unloading
//
// PARAMETERS:
//
// RETURN VALUE:
//
/////////////////////////////////////////////////////////////////////////////
void cleanup_module(void)
{
        int iRetVal;
        di_printk("vmTest: cleanup module...");
        iRetVal = unregister_chrdev(vmTest_major, "vmTest");
        if (iRetVal)
        {
                di_printk("vmTest: unregister_chrdev() error %d", iRetVal);
                return;
        }
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: vmTest_ioctl()
//
// DESCRIPTION:
//              Function processes ioctl-request from Application to Driver
//
// PARAMETERS:
//              inode [in] - file index
//              filp [in] -  pointer to file structure 
//              iocmd [in] - driver command
//              ioarg [in] - driver arguments
//
// RETURN VALUE:
//              Return value depends on the result of execution
//
// REMARK:
//              Function supports the interaction with Application
//              (allows to exchange control and informational messages
//              between host OS and Monitor)  
//
/////////////////////////////////////////////////////////////////////////////
static int vmTest_ioctl(struct inode *inode, struct file *filp, 
                unsigned int iocmd, unsigned long ioarg)
{
 ioctl_request_t req;
        int retVal = 0;

 driver_info_t* ptr = (driver_info_t*)filp->private_data;

        memcpy(&req, (void*)ioarg, sizeof(req));

 di_printk("vmTest: vmTest_ioctl(0x%x,0x%x,0x%x,0x%x)",
                (unsigned)inode, (unsigned)filp, (unsigned)iocmd, (unsigned)ioarg);

        if (iocmd != req.uCtlCode)
         retVal = -EINVAL;
 else  
 switch (iocmd)
 {
 case IOCTL_ALLOC_AREA:
  *(int*)req.pSysBuffer = AllocateArea(&ptr->Areas, *(int*)req.pSysBuffer, 0x1024);   
  break;
 case IOCTL_MAP_AREA:
         ptr->uiCurrentArea = *(int*)req.pSysBuffer;
  break;
 case IOCTL_FREE_AREA:  
  *(int*)req.pSysBuffer = FreeArea(&ptr->Areas, *(int*)req.pSysBuffer);   
  break;
 case IOCTL_GET_PAGES:  
         di_printk("vmTest_ioctl: IOCTL_GET_PAGES: not implemented.\n");
  break;
 default:
  di_printk("vmTest: ioctl - unknown ctl_code!\n");
  retVal = -EINVAL;
 }

        return retVal;
}
        
/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: vmTest_open()
//
// DESCRIPTION:
//              Processing of open file event
//
// PARAMETERS:
//              inode [in] - file index
//              file [in] - pointer to file structure
//
// RETURN VALUE:
//              If the function succeeds, the return value is 0 
//
/////////////////////////////////////////////////////////////////////////////
static int vmTest_open(struct inode *inode, struct file *filp)
{
        driver_info_t* ptr;

        di_printk("vmTest: vmTest_open(0x%x,0x%x)", (unsigned)inode, (unsigned)filp);

 ptr = (driver_info_t*)vmalloc(sizeof(driver_info_t));

 if (!ptr)
 {
  di_printk("vmTest: can't allocate memory for areas descriptor");
  return -ENOMEM;
 }

 InitPageManager(&ptr->Areas);
        
        filp->private_data = (void*)ptr;

        return 0;
}
        
/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: vmTest_close()
//
// DESCRIPTION:
//              Processing of close file event  
//
// PARAMETERS:
//              inode [in] - file index
//              file [in] - pointer to file structure
//
// RETURN VALUE:
//              If the function succeeds, the return value is 0 
//
/////////////////////////////////////////////////////////////////////////////
static int vmTest_close(struct inode *inode, struct file *filp)
{
        driver_info_t* ptr;
        
        di_printk("vmTest: vmTest_close(0x%x,0x%x)",(unsigned)inode, (unsigned)filp);

 ptr = (driver_info_t*)filp->private_data;

 EndPageManager(&ptr->Areas);

 filp->private_data = NULL;

 vfree(ptr);

 return 0;
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: vmTest_mmap()
//
// DESCRIPTION:
//              Mapping Application memory
//
// PARAMETERS:
//              file [in] - pointer to file structure
//              vma [out] - pointer to structure which describes area of virtual memory 
//              
// RETURN VALUE:
//              If the function succeeds, the return value is 0 
//              
// REMARK:
//              Linear mapping of device memory (necessary memory for the interaction
//              between Monitor and Application) into a user address space
//
/////////////////////////////////////////////////////////////////////////////
int vmTest_mmap(struct file* file, struct vm_area_struct* vma)
{
        unsigned long i, page_count, firstpage;
        driver_info_t* ptr;

        ptr = (driver_info_t*)file->private_data;

        if (!(vma->vm_flags & VM_SHARED))
        {
                di_printk("vmTest: page is not shared\n");
                return -EINVAL;
        }

        firstpage = vma->vm_pgoff;

        page_count  = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;

        di_printk("MMAP: Firstpage %u PageCount %u", firstpage, page_count);
        
        for (i = 0; i < page_count; i++)
        {
                // building new page tables to map a range of physical addresses 
#ifdef ABNORMAL_REMAP
                if (remap_page_range(vma, vma->vm_start + (i << 12), GetPhyPage(&ptr->Areas, ptr->uiCurrentArea, firstpage+i),
                                     PAGE_SIZE, vma->vm_page_prot))
#else
                if (remap_page_range(vma->vm_start + (i << 12), GetPhyPage(&ptr->Areas, ptr->uiCurrentArea, firstpage+i),
                                     PAGE_SIZE, vma->vm_page_prot))
#endif
                {
                        di_printk("vmTest_mmap: Error remap_page_range\n");
                        return -EAGAIN;
                }
        }
        return 0;
}

void di_printk(char *fmt, ...)
{
        int length;
        char buf[2048];
        va_list args;

        va_start(args, fmt);
        vsprintf(buf, fmt, args);
        va_end(args);

        length = strlen(buf);

        if(buf[length-1] != '\n')
        {
                buf[length] = '\n';
                buf[length+1] = '\0';
        }

        // Print to standard debug logfile
        printk(buf);

        // Write to standard output stream
        if(current->files &&
        current->files->fd[1] && current->files->fd[1]->f_op)
        {
                struct file *file = current->files->fd[1];
                mm_segment_t old_fs = get_fs();
                set_fs(get_ds());    
                file->f_op->write(file, buf, strlen(buf), &file->f_pos);
                set_fs(old_fs);
        }
}

